import os
import random
import pyfastx
import pyfaidx
import unittest

join = os.path.join
data_dir = join(os.path.dirname(os.path.abspath(__file__)), 'data')

gzip_fasta = join(data_dir, 'test.fa.gz')
flat_fasta = join(data_dir, 'test.fa')

class SequenceTest(unittest.TestCase):
	def setUp(self):
		self.fastx = pyfastx.Fasta(gzip_fasta)

		self.fasta = pyfastx.Fasta(flat_fasta)

		self.faidx = pyfaidx.Fasta(flat_fasta, sequence_always_upper=True)

		self.count = len(self.fastx)

	def tearDown(self):
		del self.fastx
		del self.fasta
		del self.faidx

		if os.path.exists('{}.fxi'.format(gzip_fasta)):
			os.remove('{}.fxi'.format(gzip_fasta))

		if os.path.exists('{}.fxi'.format(flat_fasta)):
			os.remove('{}.fxi'.format(flat_fasta))

		if os.path.exists('{}.fai'.format(flat_fasta)):
			os.remove('{}.fai'.format(flat_fasta))

	def get_random_index(self):
		return random.randint(0, self.count-1)

	def test_seq_by_index(self):
		#test get seq by index
		idx = self.get_random_index()
		expect = self.faidx[idx][:]
		result = self.fastx[idx]

		self.assertEqual(expect.name, result.name)
		self.assertEqual(expect.seq, result.seq)

		#test subseq
		self.assertEqual(expect[0:10].seq, result[0:10].seq)

		#test negative index
		idx = (self.get_random_index() + 1) * -1
		expect = self.faidx[idx][:]
		result = self.fastx[idx]

		self.assertEqual(expect.name, result.name)
		self.assertEqual(expect.seq, result.seq)

	def test_seq_by_key(self):
		idx = self.get_random_index()
		key = list(self.faidx.keys())[idx]

		expect = self.faidx[key][:]
		result = self.fastx[key]

		self.assertEqual(expect.name, result.name)
		self.assertEqual(expect.seq, result.seq)

	def test_seq_reverse_complement(self):
		idx = self.get_random_index()
		expect = self.faidx[idx][:]
		result = self.fastx[idx]

		self.assertEqual(str(expect.reverse), result.reverse)
		self.assertEqual(str(expect.complement), result.complement)
		self.assertEqual(str(-expect), result.antisense)

	def test_seq_raw(self):
		idx = self.get_random_index()

		seq = self.fastx[idx]

		lines = []
		with open(flat_fasta) as fh:
			for line in fh:
				if line.startswith('>{}'.format(seq.name)):
					lines.append(line)
					continue

				if lines:
					if line[0] == '>':
						break

					lines.append(line)

		self.assertEqual(''.join(lines).replace('\n','\r\n'), seq.raw)

	def test_seq_slice(self):
		idx = self.get_random_index()
		expect = str(self.faidx[idx])
		result = self.fastx[idx]
		flatseq = self.fasta[idx]

		#test gzip subseq
		self.assertEqual(expect[5:10], result[5:10].seq)

		#test flat subseq
		self.assertEqual(expect[5:10], flatseq[5:10].seq)
		self.assertEqual(expect[20:], result[20:].seq)

		#test two level slice
		self.assertEqual(expect[10:100][:20], result[10:100][:20].seq)

		#test sequence index
		pos = random.randint(0, len(expect) - 1)
		self.assertEqual(expect[pos], result[pos])

		pos = random.randint(1, len(expect)) * -1
		self.assertEqual(expect[pos], result[pos])

		del flatseq

	def test_seq_content(self):
		idx = self.get_random_index()
		result = self.fastx[idx]
		expect = self.faidx[idx]

		content = {'A': 0, 'C': 0, 'G': 0, 'T': 0}
		content['A'] += expect[:].seq.count('A')
		content['C'] += expect[:].seq.count('C')
		content['G'] += expect[:].seq.count('G')
		content['T'] += expect[:].seq.count('T')

		expect_gc = (content['G']+content['C'])/sum(content.values())*100

		self.assertEqual(result.composition, content)
		self.assertEqual(round(result.gc_content, 3), round(expect_gc, 3))

		#test gc skew
		expect_skew = (content['G']-content['C'])/(content['G']+content['C'])
		self.assertEqual(round(result.gc_skew, 3), round(expect_skew, 3))

	def test_seq_iter(self):
		idx = self.get_random_index()
		fai_seq = self.faidx[idx]
		fxi_seq = self.fastx[idx]
		fas_seq = self.fasta[idx]

		# test read seq line by line
		flatsq = [line for line in fas_seq]
		expect = [str(line) for line in fai_seq]
		result = [line for line in fxi_seq]

		self.assertEqual(expect, result)
		self.assertEqual(expect, flatsq)

		# test seq long name
		self.assertEqual(fai_seq.long_name.strip(), fxi_seq.description)

		# test seq str
		self.assertEqual(str(fai_seq), str(fxi_seq))

		# test seq contains
		s, e = sorted(random.sample(range(1, len(fai_seq)), 2))
		segment = str(fai_seq)[s-1:e]
		self.assertTrue(segment in fxi_seq)

		# test seq search
		expect = str(fai_seq).index(segment) + 1
		result = fxi_seq.search(segment)
		self.assertEqual(expect, result)

	def test_seq_repr(self):
		s = self.fastx[0]
		n = s.name
		self.assertEqual(repr(s), "<Sequence> {} with length of {}".format(s.name, len(s)))

		s = self.fastx[0][0:11]
		self.assertEqual(repr(s), "<Sequence> {} from {} to {}".format(n, s.start, s.end))

		#test name
		self.assertEqual(s.name, "{}:{}-{}".format(n, s.start, s.end))

	def test_full_compo(self):
		idx = self.get_random_index()
		fas = pyfastx.Fasta(flat_fasta, full_index=True)
		self.assertEqual(self.fastx[idx].composition, fas[idx].composition)

	def test_seq_exception(self):
		with self.assertRaises(RuntimeError):
			for line in self.fastx[0][10:20]:
				print(line)

if __name__ == '__main__':
	unittest.main()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import os

project = 'pyfastx'
copyright = '2023, Lianming Du'
author = 'Lianming Du'

with open(os.path.join(os.path.abspath('..'), 'src', 'version.h')) as fh:
	__version__ = fh.read().split()[2].strip('"')

version = __version__

release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

#html_theme = 'alabaster'
html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



import os
import random
import pyfastx
import pyfaidx
import unittest

join = os.path.join
root_dir = join(os.path.dirname(os.path.abspath(__file__)), '..')
data_dir = join(os.path.dirname(os.path.abspath(__file__)), 'data')

gzip_fasta = join(data_dir, 'test.fa.gz')
flat_fasta = join(data_dir, 'test.fa')
rna_fasta = join(data_dir, 'rna.fa')
protein_fasta = join(data_dir, 'protein.fa')


class FastaTest(unittest.TestCase):
	def setUp(self):
		self.fastx = pyfastx.Fasta(gzip_fasta)

		self.fasta = pyfastx.Fasta(flat_fasta)

		self.faidx = pyfaidx.Fasta(flat_fasta, sequence_always_upper=True)

		self.count = len(self.fastx)

	def tearDown(self):
		#fix permission error on Windows
		del self.fastx
		del self.fasta
		del self.faidx

		if os.path.exists('{}.fxi'.format(gzip_fasta)):
			os.remove('{}.fxi'.format(gzip_fasta))

		if os.path.exists('{}.fxi'.format(flat_fasta)):
			os.remove('{}.fxi'.format(flat_fasta))

		if os.path.exists('{}.fai'.format(flat_fasta)):
			os.remove('{}.fai'.format(flat_fasta))

		if os.path.exists('{}.fxi'.format(rna_fasta)):
			os.remove('{}.fxi'.format(rna_fasta))

		if os.path.exists('{}.fxi'.format(protein_fasta)):
			os.remove('{}.fxi'.format(protein_fasta))

	def get_random_index(self):
		return random.randint(0, self.count-1)

	def test_module(self):
		# gzip check test
		self.assertEqual(pyfastx.gzip_check(gzip_fasta), self.fastx.is_gzip)

		# version test
		with open(join(root_dir, 'src', 'version.h')) as fh:
			version = fh.read().split()[2].strip('"')
			self.assertEqual(version, pyfastx.version())

		#reverse complement
		self.assertEqual(pyfastx.reverse_complement('ATGC'), 'GCAT')

	def test_build(self):
		self.fastx = pyfastx.Fasta(gzip_fasta, build_index=False)

		if os.path.exists('{}.fxi'.format(gzip_fasta)):
			os.remove('{}.fxi'.format(gzip_fasta))

		self.fastx.build_index()

	def test_fasta(self):
		#test gzip
		self.assertFalse(self.fasta.is_gzip)

		#seq counts
		self.assertEqual(len(self.fastx), len(self.faidx.keys()))

		#seq length
		expect_size = sum(len(s) for s in self.faidx)
		self.assertEqual(self.fastx.size, expect_size)

		#test composition
		expect = {'A': 0, 'C': 0, 'G': 0, 'T': 0}
		for s in self.faidx:
			expect['A'] += s[:].seq.count('A')
			expect['C'] += s[:].seq.count('C')
			expect['G'] += s[:].seq.count('G')
			expect['T'] += s[:].seq.count('T')
		self.assertEqual(self.fastx.composition, expect)

		#test GC content
		expect_gc = (expect['G']+expect['C'])/sum(expect.values())*100
		self.assertEqual(round(self.fastx.gc_content, 3), round(expect_gc, 3))

		#test GC skew
		expect_skew = (expect['G']-expect['C'])/(expect['G']+expect['C'])
		self.assertEqual(round(self.fastx.gc_skew, 3), round(expect_skew, 3))

		#test longest and shortest sequence
		longest = (None, 0)
		shortest = (None, expect_size)
		for seq in self.faidx:
			l = len(seq)
			if l > longest[1]:
				longest = (seq.name, l)

			if l < shortest[1]:
				shortest = (seq.name, l)

		long_seq = self.fastx.longest
		short_seq = self.fastx.shortest

		self.assertEqual(longest, (long_seq.name, len(long_seq)))
		self.assertEqual(shortest, (short_seq.name, len(short_seq)))

		#test contains
		idx = self.get_random_index()
		name = self.faidx[idx].name
		self.assertTrue(name in self.fastx)

	#test repr
	def test_repr(self):
		expect = "<Fasta> {} contains {} sequences".format(gzip_fasta, self.count)
		result = repr(self.fastx)
		self.assertEqual(expect, result)

		#without build index
		fa = pyfastx.Fasta(flat_fasta, build_index=False)
		expect = "<Fasta> {}".format(flat_fasta)
		result = repr(fa)
		self.assertEqual(expect, result)

	def test_seq_type(self):
		#test dna format
		self.assertEqual(self.fastx.type, 'DNA')

		#test rna format
		rna = pyfastx.Fasta(rna_fasta)
		self.assertEqual(rna.type, "RNA")

		#test protein format
		prot = pyfastx.Fasta(protein_fasta)
		self.assertEqual(prot.type, "protein")

	def test_iter_object(self):
		for seq in self.fastx:
			expect = self.faidx[seq.name][:].seq
			self.assertEqual(expect, seq.seq)

		for seq in pyfastx.Fasta(flat_fasta, uppercase=True):
			expect = self.faidx[seq.name][:].seq
			self.assertEqual(expect, seq.seq)

		#test reference of sequence made from loop
		for seq in self.fastx:
			break

		expect = self.faidx[seq.name][:].seq
		self.assertEqual(expect, seq.seq)
		self.assertEqual(expect, seq.seq)

	def test_iter_tuple(self):
		fa = pyfastx.Fasta(gzip_fasta, build_index=False)
		
		for name, seq in fa:
			expect = str(self.faidx[name])
			self.assertEqual(expect, seq)

	def test_iter_upper(self):
		fa = pyfastx.Fasta(flat_fasta, build_index=False, uppercase=True)

		for name, seq in fa:
			self.assertEqual(seq, self.faidx[name.split()[0]][:].seq)

	def test_iter_full_name(self):
		fa = pyfastx.Fasta(flat_fasta, build_index=False, full_name=True)

		for name, _ in fa:
			self.assertTrue(name, self.fastx[name.split()[0]].description)

	def test_iter_upper_full_name(self):
		fa = pyfastx.Fasta(flat_fasta, build_index=False, uppercase=True, full_name=True)

		for name, seq in fa:
			self.assertEqual(name, self.fastx[name.split()[0]].description)
			self.assertEqual(seq, self.fastx[name.split()[0]].seq)

	def test_key_func(self):
		del self.fastx

		#remove previously created index file
		if os.path.exists("{}.fxi".format(gzip_fasta)):
			os.remove("{}.fxi".format(gzip_fasta))

		self.fastx = pyfastx.Fasta(gzip_fasta, key_func=lambda x: x.split()[1])
		idx = self.get_random_index()
		self.assertEqual(self.fastx[idx].name, self.fastx[idx].description.split()[1])

	def test_statistics(self):
		lens = sorted([len(seq) for seq in self.faidx], reverse=True)
		half = sum(lens)/2
		tlen = 0
		l50 = 0
		for n50 in lens:
			l50 += 1
			tlen += n50

			if tlen >= half:
				break

		self.assertEqual(self.fastx.nl(50), (n50, l50))

		#test mean length
		expect = round(sum(lens)/len(lens), 3)
		result = round(self.fastx.mean, 3)
		self.assertEqual(expect, result)

		#test median length
		lens = sorted(lens)
		expect = lens[105]

		result = self.fastx.median
		self.assertEqual(expect, result)

		#test count squence
		expect = 0
		for l in lens:
			if l >= 200:
				expect += 1
		result = self.fastx.count(200)
		self.assertEqual(expect, result)

	def test_seq_fetch(self):
		idx = self.get_random_index()
		name = list(self.faidx.keys())[idx]
		l = len(self.fastx[idx])

		#test one interval
		a = int(l/2)
		interval = (random.randint(1, a), random.randint(a+1, l))

		expect = str(self.faidx[name])[interval[0]-1:interval[1]]
		result = self.fastx.fetch(name, interval)

		self.assertEqual(expect, result)

		#test multiple intervals
		intervals = []
		intervals.append((random.randint(1, int(a/2)), random.randint(int(a/2)+1, a)))
		intervals.append((random.randint(a+1, int((a+l)/2)), random.randint(int((a+l)/2)+1, l)))

		expect = "".join([str(self.faidx[name])[s-1:e] for s, e in intervals])
		result = self.fastx.fetch(name, intervals)

		self.assertEqual(expect, result)

	def test_seq_flank(self):
		idx = self.get_random_index()
		name = list(self.faidx.keys())[idx]
		l = len(self.fastx[idx])

		a = int(l/2)
		start = random.randint(1, a)
		end = random.randint(a+1, l)
		flen = 20
		left, right = self.fastx.flank(name, start, end, flen)
		s = start - flen - 1
		if s < 0:
			s = 0

		self.assertEqual(str(self.faidx[name])[s:start-1], left)
		self.assertEqual(str(self.faidx[name])[end:end+flen], right)

		left, right = self.fastx.flank(name, start, end, flank_length=flen, use_cache=True)
		self.assertEqual(str(self.faidx[name])[s:start-1], left)
		self.assertEqual(str(self.faidx[name])[end:end+flen], right)

		left, right = self.fastx.flank(name, 1, len(self.faidx[name]))
		self.assertEqual('', left)
		self.assertEqual('', right)

	def test_no_upper(self):
		fa = pyfastx.Fasta(flat_fasta, uppercase=False)
		self.assertEqual(self.fastx[0].seq, fa[0].seq)

	def test_exception(self):
		with self.assertRaises(TypeError):
			pyfastx.Fasta(flat_fasta, key_func=1)

		with self.assertRaises(FileExistsError):
			pyfastx.Fasta('a_file_not_exists')

		with self.assertRaises(ValueError):
			self.fastx.fetch('seq1', {'a':1})

		with self.assertRaises(NameError):
			self.fastx.fetch('seq1', (1,10))

		with self.assertRaises(ValueError):
			self.fastx.fetch(self.fastx[0].name, (1,10,20))

		with self.assertRaises(ValueError):
			self.fastx.fetch(self.fastx[0].name, (20, 10))

		with self.assertRaises(ValueError):
			self.fastx.fetch(self.fastx[0].name, [20, 10])

		with self.assertRaises(IndexError):
			_ = self.fastx[self.count]

		with self.assertRaises(KeyError):
			_ = self.fastx[list()]

		with self.assertRaises(ValueError):
			self.fastx.nl(101)

		with self.assertRaises(RuntimeError):
			non_fa = 'non.fa'
			with open(non_fa, 'w') as fw:
				fw.write('abc')

			pyfastx.Fasta(non_fa)

			os.remove(non_fa)

if __name__ == '__main__':
	unittest.main()



"""randomly get numbers of sequence name from fasta file
Usage:
	python3 get_random_sequence_names.py seed prop fastas...
@seed, random seed
@prop, proportion of sequence to extract, 0-1
"""

import sys
import math
import random
import pyfastx

random.seed(sys.argv[1])
prop = float(sys.argv[2])

for fafile in sys.argv[3:]:
	fa = pyfastx.Fasta(fafile)
	num = math.ceil(len(fa)*prop)
	ids = fa.keys()

	samples = random.sample(range(len(fa)), num)

	with open("{}.list".format(fafile), 'w') as fw:
		for i in samples:
			fw.write("{}\n".format(ids[i]))



import os
import random
import pyfastx
import unittest

join = os.path.join
data_dir = join(os.path.dirname(os.path.abspath(__file__)), 'data')

gzip_fastq = join(data_dir, 'test.fq.gz')
flat_fastq = join(data_dir, 'test.fq')

class FastqTest(unittest.TestCase):
	def setUp(self):

		self.fastq = pyfastx.Fastq(gzip_fastq)

		#reload index
		self.fastq = pyfastx.Fastq(gzip_fastq)

		#flat fastq
		self.flatq = pyfastx.Fastq(flat_fastq)

		self.reads = {}
		self.bases = {'A': 0, 'T': 0, 'G': 0, 'C':0, 'N':0}
		i = 0
		c = -1
		with open(flat_fastq) as fh:
			for line in fh:
				i += 1

				if i % 4 == 1:
					c += 1
					self.reads[c] = [line[1:].strip().split()[0], 0, 0]

				elif i % 4 == 2:
					self.reads[c][1] = line.strip()

					self.bases['A'] += line.count('A')
					self.bases['T'] += line.count('T')
					self.bases['G'] += line.count('G')
					self.bases['C'] += line.count('C')
					self.bases['N'] += line.count('N')

				elif i % 4 == 0:
					self.reads[c][2] = line.strip()

	def tearDown(self):
		del self.fastq
		del self.flatq

		if os.path.exists('{}.fxi'.format(gzip_fastq)):
			os.remove('{}.fxi'.format(gzip_fastq))

		if os.path.exists('{}.fxi'.format(flat_fastq)):
			os.remove('{}.fxi'.format(flat_fastq))

	def get_random_read(self):
		return random.randint(0, len(self.fastq)-1)

	def test_build(self):
		del self.fastq

		if os.path.exists('{}.fxi'.format(gzip_fastq)):
			os.remove('{}.fxi'.format(gzip_fastq))

		fq = pyfastx.Fastq(gzip_fastq, build_index=False)
		fq.build_index()

		self.fastq = pyfastx.Fastq(gzip_fastq)

	def test_fastq(self):
		# test gzip format
		self.assertEqual(pyfastx.gzip_check(gzip_fastq), self.fastq.is_gzip)

		# test seq length
		self.assertEqual(self.fastq.size, sum(self.bases.values()))

		# test length
		self.assertEqual(len(self.reads), len(self.fastq))

		# test gc content
		result = round(self.fastq.gc_content, 2)

		expect = round((self.bases['G']+self.bases['C'])/(sum(self.bases.values()))*100, 2)
		self.assertEqual(expect, result)

		# test composition
		self.assertEqual(self.fastq.composition, self.bases)

		# test encoding type
		self.assertEqual(['Sanger Phred+33', 'Illumina 1.8+ Phred+33', 'PacBio HiFi Phred+33'], self.fastq.encoding_type)

		# test phred
		self.assertEqual(self.fastq.phred, 33)

		#test length
		self.assertEqual(self.fastq.avglen, 150)
		self.assertEqual(self.fastq.minlen, 150)
		self.assertEqual(self.fastq.maxlen, 150)

		#test quality
		self.assertEqual(self.fastq.maxqual, 70)
		self.assertEqual(self.fastq.minqual, 35)

	def test_platform(self):
		#test unknown
		fqfile = 'test1.fq'
		with open(fqfile, 'w') as fw:
			fw.write("@read1\n")
			fw.write("AAAAAAAAAA\n")
			fw.write("+\n")
			fw.write("          \n")
		fq = pyfastx.Fastq(fqfile)
		self.assertEqual(fq.encoding_type, ['Unknown'])
		del fq
		os.remove("{}.fxi".format(fqfile))

		with open(fqfile, 'w') as fw:
			fw.write("@read1\n")
			qs = []
			for i in range(59, 105):
				qs.append(chr(i))
			fw.write("{}\n".format('A'*len(qs)))
			fw.write("+\n")
			fw.write("{}\n".format(''.join(qs)))
		fq = pyfastx.Fastq(fqfile)
		self.assertIn("Solexa Solexa+64", fq.encoding_type)
		del fq
		os.remove("{}.fxi".format(fqfile))

		with open(fqfile, 'w') as fw:
			fw.write("@read1\n")
			qs = []
			for i in range(64, 105):
				qs.append(chr(i))
			fw.write("{}\n".format('A'*len(qs)))
			fw.write("+\n")
			fw.write("{}\n".format(''.join(qs)))
		fq = pyfastx.Fastq(fqfile)
		self.assertIn("Illumina 1.3+ Phred+64", fq.encoding_type)
		del fq
		os.remove("{}.fxi".format(fqfile))

		with open(fqfile, 'w') as fw:
			fw.write("@read1\n")
			qs = []
			for i in range(66, 105):
				qs.append(chr(i))
			fw.write("{}\n".format('A'*len(qs)))
			fw.write("+\n")
			fw.write("{}\n".format(''.join(qs)))
		fq = pyfastx.Fastq(fqfile)
		self.assertIn("Illumina 1.5+ Phred+64", fq.encoding_type)
		del fq
		os.remove("{}.fxi".format(fqfile))
		os.remove(fqfile)

	def test_negative(self):
		read = self.fastq[-1]
		self.assertEqual(read.name, self.reads[len(self.reads)-1][0])

	def test_iter_object(self):
		# test read iter
		i = -1
		for read in self.fastq:
			i += 1
			self.assertEqual(read.name, self.reads[i][0])
			self.assertEqual(read.seq, self.reads[i][1])
			self.assertEqual(read.qual, self.reads[i][2])

		#test reference of read made from loop
		for read in self.fastq:
			break

		expect = self.reads[0][1]
		self.assertEqual(expect, read.seq)
		self.assertEqual(expect, read.seq)

	def test_iter_tuple(self):
		i = -1
		for name, seq, qual in pyfastx.Fastq(flat_fastq, build_index=False):
			i += 1
			self.assertEqual(name, self.reads[i][0])
			self.assertEqual(seq, self.reads[i][1])
			self.assertEqual(qual, self.reads[i][2])

	def test_read_len(self):
		lens = [len(it[1]) for it in self.reads.values()]

		self.assertEqual(self.fastq.minlen, min(lens))
		self.assertEqual(self.fastq.maxlen, min(lens))

	def test_repr(self):
		self.assertEqual(repr(self.fastq), "<Fastq> {} contains {} reads".format(gzip_fastq, len(self.reads)))

	def test_full_name(self):
		fq = pyfastx.Fastq(flat_fastq, build_index=False, full_name=True)

		for name, _, _ in fq:
			self.assertTrue(name, self.fastq[name.split()[0]].description)

	def test_exception(self):
		with self.assertRaises(FileExistsError):
			_ = pyfastx.Fastq('a_fastq_file_not_exists')

		with self.assertRaises(IndexError):
			_ = self.fastq[len(self.fastq)]

		with self.assertRaises(KeyError):
			_ = self.fastq[int]

		with self.assertRaises(KeyError):
			_ = self.fastq['abc']

if __name__ == '__main__':
	unittest.main()



import sys
import pysam

idfile, fafile = sys.argv[1:]

fa = pysam.FastaFile(fafile)

with open(idfile) as fh:
	for line in fh:
		seqid, start, end = line.strip().split()
		s = fa.fetch(seqid, int(start), int(end))
		print(s)



import sys
import csv

mapping = {
	'GCA_002915635.1': "Axolotl",
	'GCA_001447015.2': "Sugar pine",
	'GCA_000411955.5': "White spruce",
	'GCA_900519105.1': "Bread wheat",
	'GCA_900184675.1': "Wild wheat",
	'GCA_002284835.2': "Bullfrog",
	'GCA_000516895.1': "Migratory locust",
	'GCA_003013575.1': "Pea",
	'GCF_000001405.38': "Human",
	'GCF_000005005.2': "Maize",
	'GCF_000002315.5': "Chicken",
	'GCF_000003195.3': "Sorghum",
	'GCF_000331145.1': "Chickpea",
	'GCF_000001735.4': "Thale cress",
	'GCF_000146045.2': "Yeast",
	'SRR6649487_10': '10%',
	'SRR6649487_50': '50%',
	'SRR6649487_1': '100%'
}

title = ["genome", "size", "count", "tool", "memory", "time"]

def kb2mb(v):
	return str(float(v)/1024)

def bp2gb(v):
	return str(float(v)/1000000000)

def kb2gb(v):
	return str(float(v)/1048576)

with open(sys.argv[1]) as fh:
	reader = csv.reader(fh, delimiter='\t')
	header = next(reader)

	if not header[5]:
		title.append("index")

	print("\t".join(title))
	
	column = len(header)
	for row in reader:
		g = mapping.get(row[0], row[0])
		s = bp2gb(row[1])
		c = row[2]

		i = 3
		while (i < column):
			items = [g, s, c]
			items.append(header[i])
			items.append(kb2gb(row[i]))
			i += 1
			items.append(row[i])
			i += 1

			if not header[5]:
				items.append(kb2mb(row[i]))
				i += 1

			print("\t".join(items))



import sys
import csv

title = ["file", "fsize", "gsize", "count", "tool", "memory", "time"]

def kb2mb(v):
	return str(float(v)/1024)

def bp2gb(v):
	return str(float(v)/1000000000)

def kb2gb(v):
	return str(float(v)/1048576)

def bs2gb(v):
	return str(float(v)/1073741824)

with open(sys.argv[1]) as fh:
	reader = csv.reader(fh, delimiter='\t')
	header = next(reader)

	if not header[7]:
		title.append("index")

	print("\t".join(title))
	
	column = len(header)
	for row in reader:
		fn = row[0]
		fs = bs2gb(row[3])
		gs = bs2gb(row[4])
		c = row[2]

		i = 5
		while (i < column):
			items = [fn, fs, gs, c]
			items.append(header[i])
			items.append(kb2gb(row[i]))
			i += 1
			items.append(row[i])
			i += 1

			if not header[7]:
				items.append(kb2gb(row[i]))
				i += 1

			print("\t".join(items))



import sys
from Bio import SeqIO

idfile, fafile = sys.argv[1:]

fa = SeqIO.index_db("{}.db".format(fafile))

with open(idfile) as fh:
	for line in fh:
		seqid, start, end = line.strip().split()
		s = fa[seqid][int(start):int(end)].seq
		print(s)



import os
import random
import pyfastx
import unittest

join = os.path.join
data_dir = join(os.path.dirname(os.path.abspath(__file__)), 'data')

gzip_fastq = join(data_dir, 'test.fq.gz')
flat_fastq = join(data_dir, 'test.fq')

class FastxTest(unittest.TestCase):
	def setUp(self):
		self.fastq = pyfastx.Fastq(gzip_fastq)

		with open(flat_fastq) as fh:
			self.keys = [line.split()[0][1:] for line in fh if line[0] == '@']
		self.count = len(self.keys)

	def tearDown(self):
		del self.fastq

		if os.path.exists('{}.fxi'.format(gzip_fastq)):
			os.remove('{}.fxi'.format(gzip_fastq))

	def get_random_index(self):
		return random.randint(0, self.count-1)

	def test_fastq_key(self):
		#test for repr
		keys = self.fastq.keys()
		self.assertEqual(repr(keys), "<FastqKeys> contains {} keys".format(self.count))

		#test for lenth
		self.assertEqual(self.count, len(keys))

		#test for iter
		for i, k in enumerate(keys):
			self.assertEqual(self.keys[i], k)

		#test random access
		idx = self.get_random_index()
		self.assertEqual(self.keys[idx], keys[idx])

		#test contain
		idx = self.get_random_index()
		self.assertTrue(self.keys[idx] in keys)

		#test negative index
		neg = (self.count - idx) * -1
		self.assertEqual(self.keys[neg], keys[neg])

	def test_exception(self):
		with self.assertRaises(IndexError):
			_ = self.keys[self.count*2]



import os
import gzip
import random
import pyfastx
import unittest

join = os.path.join
data_dir = join(os.path.dirname(os.path.abspath(__file__)), 'data')

gzip_fastq = join(data_dir, 'test.fq.gz')
flat_fastq = join(data_dir, 'test.fq')

class ReadTest(unittest.TestCase):
	def setUp(self):

		self.fastq = pyfastx.Fastq(gzip_fastq)

		#reload index
		self.fastq = pyfastx.Fastq(gzip_fastq)

		#flat fastq
		self.flatq = pyfastx.Fastq(flat_fastq)

		self.reads = {}
		self.bases = {'A': 0, 'T': 0, 'G': 0, 'C':0, 'N':0}
		i = 0
		c = -1
		with open(flat_fastq) as fh:
			for line in fh:
				i += 1

				if i % 4 == 1:
					c += 1
					self.reads[c] = [line[1:].strip().split()[0], 0, 0]

				elif i % 4 == 2:
					self.reads[c][1] = line.strip()

					self.bases['A'] += line.count('A')
					self.bases['T'] += line.count('T')
					self.bases['G'] += line.count('G')
					self.bases['C'] += line.count('C')
					self.bases['N'] += line.count('N')

				elif i % 4 == 0:
					self.reads[c][2] = line.strip()

	def tearDown(self):
		del self.fastq
		del self.flatq

		if os.path.exists('{}.fxi'.format(gzip_fastq)):
			os.remove('{}.fxi'.format(gzip_fastq))

		if os.path.exists('{}.fxi'.format(flat_fastq)):
			os.remove('{}.fxi'.format(flat_fastq))

	def get_random_read(self):
		return random.randint(0, len(self.fastq)-1)

	def test_repr(self):
		idx = self.get_random_read()
		read = self.flatq[idx]
		read1 = self.reads[idx]

		self.assertEqual(repr(read), "<Read> {} with length of {}".format(read1[0], len(read1[1])))

	def test_read(self):
		idx = self.get_random_read()
		result = self.fastq[idx]
		expect = self.reads[idx]

		del result
		result = self.fastq[idx]

		read0 = self.flatq[idx]

		# test length
		self.assertEqual(len(result), len(expect[1]))

		# test name
		self.assertEqual(result.name, expect[0])

		# test str
		self.assertEqual(str(result), expect[1])

		# test seq
		self.assertEqual(result.seq, expect[1])
		self.assertEqual(read0.seq, expect[1])

		# test quality
		self.assertEqual(result.qual, expect[2])
		self.assertEqual(read0.qual, expect[2])

		# test quality integer
		self.assertEqual(result.quali, [ord(b)-33 for b in expect[2]])

		result = self.fastq[expect[0]]

		# test subscript
		self.assertEqual(result.seq, expect[1])

		# test contain
		self.assertTrue(result.name in self.fastq)

	def test_read_seq(self):
		#test reverse
		idx = self.get_random_read()
		result = self.fastq[idx]
		expect = self.reads[idx][1]
		self.assertEqual(result.reverse, expect[::-1])

		#test complement
		complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N':'N'}
		expect1 = ''.join(complement.get(b) for b in expect)
		self.assertEqual(result.complement, expect1)

		#test antisense
		expect2 = ''.join(complement.get(b) for b in reversed(expect))
		self.assertEqual(result.antisense, expect2)

	def test_read_description(self):
		idx = self.get_random_read()
		read = self.fastq[idx]

		i = -1
		with open(flat_fastq) as fh:
			for line in fh:
				if line[0] == '@':
					i += 1

					if i == idx:
						break

		self.assertEqual(line.strip(), read.description)

		read = self.flatq[idx]
		self.assertEqual(line.strip(), read.description)

	def test_read_raw(self):
		idx = self.get_random_read()
		read = self.fastq[idx]

		lines = []
		with gzip.open(gzip_fastq, 'rb') as fh:
			for line in fh:
				line = line.decode()
				if line.startswith('@{}'.format(read.name)):
					lines.append(line)
					continue

				if lines:
					if line[0] == '@':
						break

					lines.append(line)

		self.assertEqual(''.join(lines), read.raw)

		idx = self.get_random_read()
		read = self.flatq[idx]
		lines = []
		with open(flat_fastq, 'rb') as fh:
			for line in fh:
				line = line.decode()
				if line.startswith('@{}'.format(read.name)):
					lines.append(line)
					continue

				if lines:
					if line[0] == '@':
						break

					lines.append(line)

		self.assertEqual(''.join(lines), read.raw)

if __name__ == '__main__':
	unittest.main()



import sys
import pyfastx

idfile, fafile = sys.argv[1:]

fa = pyfastx.Fasta(fafile)

with open(idfile) as fh:
	for line in fh:
		seqid, start, end = line.strip().split()
		s = fa[seqid][int(start):int(end)].seq
		print(s)



import os
import re
import sys
import gzip
import math
import random
import pyfastx
import argparse

def fastx_format_check(infile):
	if pyfastx.gzip_check(infile):
		fp = gzip.open(infile, 'rt')
	else:
		fp = open(infile)

	for line in fp:
		if line.strip():
			break

	fp.close()

	if line[0] == '>':
		return 'fasta'

	elif line[0] == '@':
		return 'fastq'

	else:
		raise Exception("Input file %s is not fasta or fastq", infile)

#find max and min value index
def max_index(lst):
	return max(range(len(lst)), key=lst.__getitem__)

def min_index(lst):
	return min(range(len(lst)), key=lst.__getitem__)

def is_glob(pattern):
	if re.search(r'[\*\?]', pattern) or re.search(r'\[\!?.*\]', pattern):
		return True
	else:
		return False

def print_table(table):
	long_cols = [0] * len(table[0])

	for row in table:
		for idx, col in enumerate(row):
			l = len(str(col))
			if l > long_cols[idx]:
				long_cols[idx] = l

	for row in table:
		row = ['{:<{}}'.format(col, long_cols[idx]) if idx==0 else
		'{:>{}}'.format(col, long_cols[idx]) for idx, col in enumerate(row)]

		print("\t".join(row))

def fastx_build(args):
	for infile in args.fastx:
		if infile.endswith('.fxi'):
			continue

		fastx_type = fastx_format_check(infile)

		if fastx_type == 'fasta':
			_ = pyfastx.Fasta(infile, full_index=args.full)

		elif fastx_type == 'fastq':
			_ = pyfastx.Fastq(infile, full_index=args.full)

def fastx_info(args):
	farows = [["fileName", "seqType", "seqCounts", "totalBases", "GC%",
				"avgLen", "medianLen", "maxLen", "minLen", "N50", "L50"]]
	fqrows = [["fileName", "readCounts", "totalBases", "GC%", "avgLen", "maxLen",
				"minLen", "maxQual", "minQual", "qualEncodingSystem"]]

	for infile in args.fastx:
		if infile.endswith('.fxi'):
			continue

		fastx_type = fastx_format_check(infile)

		if fastx_type == 'fasta':
			fa = pyfastx.Fasta(infile, full_index=True)
			if fa.type in ['DNA', 'RNA']:
				gc = round(fa.gc_content, 3)
			else:
				gc = '-'

			row = [os.path.basename(infile), fa.type, len(fa), fa.size, gc, round(fa.mean,3),
					round(fa.median,3), len(fa.longest), len(fa.shortest)]
			row.extend(fa.nl())
			farows.append(row)

		elif fastx_type == 'fastq':
			fq = pyfastx.Fastq(infile, full_index=True)
			row = [os.path.basename(infile), len(fq), fq.size, round(fq.gc_content,3), round(fq.avglen,3),
					fq.maxlen, fq.minlen, fq.maxqual, fq.minqual, ",".join(fq.encoding_type)]
			fqrows.append(row)

	if len(farows) > 1:
		print_table(farows)

	if len(fqrows) > 1:
		if len(farows) > 1:
			print()
		print_table(fqrows)

def fasta_split(args):
	fa = pyfastx.Fasta(args.fastx)

	if args.seq_count:
		parts_num = math.ceil(len(fa)/args.seq_count)
	else:
		parts_num = args.file_num

	name, suffix1 = os.path.splitext(os.path.basename(args.fastx))

	if fa.is_gzip:
		name, suffix2 = os.path.splitext(name)

	digit = len(str(parts_num))
	lens = [0] * parts_num

	if args.seq_count:
		seqs = [0] * parts_num

	fhs = []

	for i in range(1, parts_num+1):
		if fa.is_gzip:
			subfile = "{}.{}{}{}".format(name, str(i).zfill(digit), suffix2, suffix1)
		else:
			subfile = "{}.{}{}".format(name, str(i).zfill(digit), suffix1)

		if args.outdir is not None:
			subfile = os.path.join(args.outdir, subfile)

		if fa.is_gzip:
			fh = gzip.open(subfile, 'wt')
		else:
			fh = open(subfile, 'w')

		fhs.append(fh)

	ids = fa.keys()
	mapping = {}
	for chrom in ids.sort('length', reverse=True):
		idx = min_index(lens)
		#fhs[idx].write(fa[chrom].raw)
		mapping[chrom] = idx
		lens[idx] += len(fa[chrom])

		if args.seq_count:
			seqs[idx] += 1

			if seqs[idx] == args.seq_count:
				lens[idx] = fa.size

	for seq in fa:
		fhs[mapping[seq.name]].write(seq.raw)

	for fh in fhs:
		fh.close()

def fastq_split(args):
	fq = pyfastx.Fastq(args.fastx)

	if args.file_num:
		seqs_num = math.ceil(len(fq)/args.file_num)
		parts_num = args.file_num
	else:
		seqs_num = args.seq_count
		parts_num = math.ceil(len(fq)/seqs_num)

	name, suffix1 = os.path.splitext(os.path.basename(args.fastx))

	if fq.is_gzip:
		name, suffix2 = os.path.splitext(name)

	digit = len(str(parts_num))

	seq_write = 0
	fh = None
	file_num = 0

	for read in fq:
		if seq_write == 0:
			file_num += 1

			if fq.is_gzip:
				subfile = "{}.{}{}{}".format(name, str(file_num).zfill(digit), suffix2, suffix1)
			else:
				subfile = "{}.{}{}".format(name, str(file_num).zfill(digit), suffix1)

			if args.outdir is not None:
				subfile = os.path.join(args.outdir, subfile)

			if fq.is_gzip:
				fh = gzip.open(subfile, 'wt')
			else:
				fh = open(subfile, 'w')

		fh.write(read.raw)
		seq_write += 1

		if seq_write == seqs_num:
			fh.close()
			seq_write = 0

	fh.close()

def fastx_split(args):
	fastx_type = fastx_format_check(args.fastx)

	if fastx_type == 'fasta':
		fasta_split(args)

	elif fastx_type == 'fastq':
		fastq_split(args)

def fastx_fq2fa(args):
	fq = pyfastx.Fastq(args.fastx, build_index=False)

	if args.out_file:
		fh = open(args.out_file, 'w')
	else:
		fh = sys.stdout

	for name, seq, _ in fq:
		fh.write(">{}\n{}\n".format(name, seq))

	if args.out_file:
		fh.close()
	else:
		fh.flush()


def fastx_subseq(args):
	fa = pyfastx.Fasta(args.fastx)

	if args.out_file:
		fw = open(args.out_file, 'w')
	else:
		fw = sys.stdout

	if args.region_file:
		with open(args.region_file) as fh:
			for line in fh:
				chrom, start, end = line.strip().split()
				start = int(start)
				end = int(end)
				seq = fa.fetch(chrom, (start, end))
				fw.write(">{}:{}-{}\n{}\n".format(chrom, start, end, seq))

	elif args.bed_file:
		with open(args.bed_file) as fh:
			for line in fh:
				chrom, start, end = line.strip().split()
				start = int(start) + 1
				end = int(end)
				seq = fa.fetch(chrom, (start, end))
				fw.write(">{}:{}-{}\n{}\n".format(chrom, start, end, seq))

	elif args.regions:
		for region in args.regions:
			chrom, start, end = re.split('[:-]', region)
			start = int(start)
			end = int(end)
			seq = fa[chrom][start-1:end].seq
			fw.write(">{}:{}-{}\n{}\n".format(chrom, start, end, seq))

	else:
		raise Exception("no regions or region file provided")

	if args.out_file:
		fw.close()
	else:
		fw.flush()

def fastx_sample(args):
	fastx_type = fastx_format_check(args.fastx)

	if fastx_type == 'fasta':
		Fastx = pyfastx.Fasta

	elif fastx_type == 'fastq':
		Fastx = pyfastx.Fastq

	else:
		raise Exception("the input file is not fasta or fastq file")

	fx = Fastx(args.fastx)

	if args.num is not None and args.num > 0:
		seq_num = args.num
		if seq_num > len(fx):
			seq_num = len(fx)

	elif args.prop is not None and 0 < args.prop <= 1:
		seq_num = math.ceil(len(fx)*args.prop)

	else:
		raise RuntimeError("specify a right seq number or proportion")

	#set random seed
	if args.seed:
		random.seed(args.seed)
	selected = random.sample(range(len(fx)), k=seq_num)

	if args.out_file is None:
		fw = sys.stdout
	else:
		fw = open(args.out_file, 'w')

	if args.sequential_read:
		writed = 0
		selected = set(selected)
		for it in fx:
			if it.id in selected:
				fw.write(it.raw)
				writed += 1
			else:
				if writed == seq_num:
					break
				_ = it.raw
	else:
		selected.sort()
		for idx in selected:
			fw.write(fx[idx].raw)

	if args.out_file is None:
		fw.flush()
	else:
		fw.close()

def fastx_extract(args):
	fastx_type = fastx_format_check(args.fastx)

	if fastx_type == 'fasta':
		Fastx = pyfastx.Fasta

	elif fastx_type == 'fastq':
		Fastx = pyfastx.Fastq

	else:
		raise Exception("the input file is not fasta or fastq file")

	fx = Fastx(args.fastx)

	if args.out_file:
		fw = open(args.out_file, 'w')
	else:
		fw = sys.stdout

	if args.list_file:
		with open(args.list_file) as fh:
			if args.sequential_read:
				names = {line.strip() for line in fh}
				total = len(names)
				writed = 0

				for it in fx:
					if it.name in names:
						fw.write(it.raw)
						writed += 1
					else:
						if writed == total:
							break
						_ = it.raw
			else:
				for line in fh:
					name = line.strip()
					fw.write(fx[name].raw)

	elif args.names:
		for name in args.names:
			fw.write(fx[name].raw)

	else:
		raise Exception("no sequence name or list file provided")

	if args.out_file:
		fw.close()
	else:
		fw.flush()

def main():
	parser = argparse.ArgumentParser(
		prog = 'pyfastx',
		usage = "pyfastx COMMAND [OPTIONS]",
		description = "A command line tool for FASTA/Q file manipulation",
		formatter_class = argparse.RawDescriptionHelpFormatter
	)

	parser.add_argument('-v', '--version',
		action = 'version',
		version = "%(prog)s version {}".format(pyfastx.version())
	)

	subparsers = parser.add_subparsers(
		title = 'Commands',
		prog = 'pyfastx',
		metavar = ''
	)

	#build index command
	parser_build = subparsers.add_parser('index',
		help = "build index for fasta/q file"
	)
	parser_build.set_defaults(func=fastx_build)
	parser_build.add_argument('-f', '--full',
		help = "build full index, base composition will be calculated",
		action = 'store_true'
	)
	parser_build.add_argument('fastx',
		help = "fasta or fastq file, gzip support",
		nargs = '+'
	)

	#statistics command
	parser_info = subparsers.add_parser('stat',
		help = "show detailed statistics information of fasta/q file"
	)
	parser_info.set_defaults(func=fastx_info)
	parser_info.add_argument('fastx',
		help = "fasta or fastq file, gzip support",
		nargs = '+'
	)

	#split command
	parser_split = subparsers.add_parser('split',
		help = "split fasta/q file into multiple files"
	)
	parser_split.set_defaults(func=fastx_split)
	split_group = parser_split.add_mutually_exclusive_group(
		required=True
	)
	split_group.add_argument('-n',
		dest = 'file_num',
		type = int,
		metavar = 'int',
		help = "split a fasta/q file into N new files with even size"
	)
	split_group.add_argument('-c',
		dest = 'seq_count',
		type = int,
		metavar = 'int',
		help = "split a fasta/q file into multiple files containing the same sequence counts"
	)
	parser_split.add_argument('-o', '--out-dir',
		dest = 'outdir',
		help = "output directory, default is current folder",
		metavar = 'str'
	)
	parser_split.add_argument('fastx',
		help = 'fasta or fastq file, gzip support'
	)

	#convert fastq to fasta command
	parser_fq2fa = subparsers.add_parser('fq2fa',
		help = "convert fastq file to fasta file"
	)
	parser_fq2fa.set_defaults(func=fastx_fq2fa)
	parser_fq2fa.add_argument('-o', '--out-file',
		metavar = 'str',
		help = "output file, default: output to stdout"
	)

	parser_fq2fa.add_argument('fastx',
		help = "fastq file, gzip support"
	)

	#get subseq from fasta
	parser_subseq = subparsers.add_parser('subseq',
		help = "get subsequences from fasta file by region"
	)
	parser_subseq.set_defaults(func=fastx_subseq)

	subseq_group = parser_subseq.add_mutually_exclusive_group()
	subseq_group.add_argument('-r', '--region-file',
		metavar = 'str',
		help = "tab-delimited file, one region per line, both start and end position are 1-based"
	)
	subseq_group.add_argument('-b', '--bed-file',
		metavar = 'str',
		help = "tab-delimited BED file, 0-based start position and 1-based end position"
	)
	parser_subseq.add_argument('-o', '--out-file',
		metavar = 'str',
		help = "output file, default: output to stdout"
	)
	parser_subseq.add_argument('fastx',
		help = "input fasta file, gzip support"
	)
	parser_subseq.add_argument('regions',
		help = "format is chr:start-end, start and end position is 1-based, multiple regions were separated by space",
		metavar = 'region',
		nargs = '*'
	)

	parser_sample = subparsers.add_parser('sample',
		help = "randomly sample sequences from fasta or fastq file"
	)
	parser_sample.set_defaults(func=fastx_sample)
	sample_group = parser_sample.add_mutually_exclusive_group(
		required = True
	)
	sample_group.add_argument('-n',
		dest = 'num',
		help = "number of sequences to be sampled",
		type = int,
		metavar = 'int'
	)
	sample_group.add_argument('-p',
		dest = 'prop',
		help = "proportion of sequences to be sampled, 0~1",
		type = float,
		metavar = 'float'
	)
	parser_sample.add_argument('-s', '--seed',
		help = "random seed, default is the current system time",
		type = int,
		default = None,
		metavar = 'int'
	)
	parser_sample.add_argument('--sequential-read',
		action = 'store_true',
		help = "start sequential reading, particularly suitable for sampling large numbers of sequences"
	)
	parser_sample.add_argument('-o', '--out-file',
		metavar = 'str',
		help = "output file, default: output to stdout"
	)
	parser_sample.add_argument('fastx',
		help = "fasta or fastq file, gzip support"
	)

	#extract sequences
	parser_extract = subparsers.add_parser('extract',
		help = "extract full sequences or reads from fasta/q file"
	)
	parser_extract.set_defaults(func=fastx_extract)

	parser_extract.add_argument('-l', '--list-file',
		metavar = 'str',
		help = "a file containing sequence or read names, one name per line"
	)
	parser_extract.add_argument('--reverse-complement',
		help = "output reverse complement sequence",
		action = 'store_true'
	)
	parser_extract.add_argument('--out-fasta',
		help = "output fasta format when extract reads from fastq, default output fastq format",
		action = 'store_true'
	)
	parser_extract.add_argument('-o', '--out-file',
		metavar = 'str',
		help = "output file, default: output to stdout"
	)
	parser_extract.add_argument('--sequential-read',
		action = 'store_true',
		help = "start sequential reading, particularly suitable for extracting large numbers of sequences"
	)
	parser_extract.add_argument('fastx',
		help = "fasta or fastq file, gzip support"
	)
	parser_extract.add_argument('names',
		metavar = 'name',
		help = "sequence name or read name, multiple names were separated by space",
		nargs = '*'
	)

	args = parser.parse_args()

	if hasattr(args, 'func'):
		args.func(args)
	else:
		parser.print_help()

if __name__ == '__main__':
	main()



import sys
import math
import random
import pyfastx

random.seed(sys.argv[1])

for fafile in sys.argv[2:]:
	fa = pyfastx.Fasta(fafile)
	ids = fa.keys()
	names = list(ids.filter(ids>1000))

	list_fh = open("{}.interval".format(fafile), 'w')
	region_fh = open("{}.region".format(fafile), 'w')

	if len(names) >= 1000:
		lists = random.sample(names, 1000)
	else:
		lists = random.sample(names * math.ceil(1000/len(names)), 1000)

	for name in lists:
		pos = random.randint(1, len(fa[name])-1000)
		list_fh.write("{}\t{}\t{}\n".format(name, pos, pos+1000))
		region_fh.write("{}:{}-{}\n".format(name, pos, pos+1000))

	list_fh.close()
	region_fh.close()



import os
import pyfaidx
import pyfastx
import unittest

join = os.path.join
data_dir = join(os.path.dirname(os.path.abspath(__file__)), 'data')

gzip_fasta = join(data_dir, 'test.fa.gz')
gzip_fastq = join(data_dir, 'test.fq.gz')
flat_fasta = join(data_dir, 'test.fa')
flat_fastq = join(data_dir, 'test.fq')

class FastxTest(unittest.TestCase):
	def setUp(self):
		self.faidx = pyfaidx.Fasta(flat_fasta)

	def tearDown(self):
		del self.faidx

		if os.path.exists('{}.fai'.format(flat_fasta)):
			os.remove('{}.fai'.format(flat_fasta))

	def test_fasta_iter(self):
		for name, seq, comment in pyfastx.Fastx(gzip_fasta, comment=True):
			s = self.faidx[name]
			self.assertEqual(str(seq), seq)
			self.assertEqual(' '.join(s.long_name.split()[1:]), comment)

	def test_fasta_upper(self):
		for name, seq in pyfastx.Fastx(flat_fasta, uppercase=True):
			self.assertEqual(str(self.faidx[name]), seq)

	def test_fastq_iter(self):
		reads = {}
		with open(flat_fastq) as fh:
			for line in fh:
				line = line.strip()

				if line[0] == '@':
					name = line.split()[0][1:]
					comment = ' '.join(line.split()[1:])
					reads[name] = [comment]
				elif line == '+':
					continue
				else:
					reads[name].append(line)

		for name, seq, qual, comment in pyfastx.Fastx(gzip_fastq, "fastq", comment=True):
			r = reads[name]

			self.assertEqual(r[0], comment)
			self.assertEqual(r[1], seq)
			self.assertEqual(r[2], qual)

	def test_fastx_repr(self):
		fa = pyfastx.Fastx(gzip_fasta, "fasta")
		self.assertEqual(repr(fa), "<Fastx> fasta {}".format(gzip_fasta))

		fq = pyfastx.Fastx(gzip_fastq, "fastq")
		self.assertEqual(repr(fq), "<Fastx> fastq {}".format(gzip_fastq))

	def test_exception(self):
		with self.assertRaises(FileExistsError):
			_ = pyfastx.Fastx('test_file')

		with self.assertRaises(RuntimeError):
			_ = pyfastx.Fastx(gzip_fasta, format="fastx")



import os
import random
import pyfastx
import pyfaidx
import unittest

join = os.path.join
data_dir = join(os.path.dirname(os.path.abspath(__file__)), 'data')

flat_fasta = join(data_dir, 'test.fa')

class SequenceErrorTest(unittest.TestCase):
	def setUp(self):
		self.fastx = pyfastx.Fasta(flat_fasta)

		self.faidx = pyfaidx.Fasta(flat_fasta, sequence_always_upper=True)

		self.count = len(self.fastx)

	def tearDown(self):
		del self.fastx
		del self.faidx

		if os.path.exists('{}.fxi'.format(flat_fasta)):
			os.remove('{}.fxi'.format(flat_fasta))

		if os.path.exists('{}.fai'.format(flat_fasta)):
			os.remove('{}.fai'.format(flat_fasta))

	def get_random_index(self):
		return random.randint(0, self.count-1)

	def get_random_key(self):
		idx = self.get_random_index()
		return list(self.faidx.keys())[idx]

	def test_seq_by_index(self):
		# get valid index
		idx = self.get_random_index()

		expect = self.faidx[idx][:]
		result = self.fastx[idx]

		self.assertEqual(expect.name, result.name)
		self.assertEqual(expect.seq, result.seq)

		# get invalid index
		with self.assertRaises(IndexError):
		    self.fastx[self.count + 100]

		# get valid index again
		expect = self.faidx[idx][:]
		result = self.fastx[idx]

		self.assertEqual(expect.name, result.name)
		self.assertEqual(expect.seq, result.seq)

	def test_seq_by_key(self):
		# get valid key
		key = self.get_random_key()

		expect = self.faidx[key][:]
		result = self.fastx[key]

		self.assertEqual(expect.name, result.name)
		self.assertEqual(expect.seq, result.seq)

		# get invalid key
		with self.assertRaises(KeyError):
		    self.fastx['__BOGUS_KEY_FOR_TESTING__']

		# get valid key again
		expect = self.faidx[key][:]
		result = self.fastx[key]

		self.assertEqual(expect.name, result.name)
		self.assertEqual(expect.seq, result.seq)


if __name__ == '__main__':
	unittest.main()



import sys
import random
import pyfastx

random.seed(sys.argv[1])

for fqfile in sys.argv[2:]:
	fq = pyfastx.Fastq(fqfile)
	samples = set(random.sample(range(len(fq)), 10000))
	with open("{}.list".format(fqfile), 'w') as fw:
		for r in fq:
			if (r.id - 1) in samples: 
				fw.write("{}\n".format(r.name))

	print(fqfile)
	



import os
import sys
import glob
import tarfile
import zipfile
import urllib.request
from setuptools import setup, Extension
from setuptools.command import build_ext

root_dir = os.path.dirname(os.path.abspath(__file__))

sources = glob.glob(os.path.join(root_dir, 'src', '*.c'))
link_args = []
comp_args = []
include_dirs = []

def prepare_zlib():
    global include_dirs
    global sources

    zlib_dir = os.path.join(root_dir, "zlib-1.3.1")
    zlib_file = os.path.join(root_dir, "zlib-1.3.1.zip")
    url = "https://github.com/madler/zlib/releases/download/v1.3.1/zlib131.zip"

    if not os.path.exists(zlib_dir):
        if not os.path.isfile(zlib_file):
            urllib.request.urlretrieve(url, zlib_file)

        with zipfile.ZipFile(zlib_file) as _zip:
            _zip.extractall()

    include_dirs.append(zlib_dir)
    sources.extend(glob.glob(os.path.join(zlib_dir, '*.c')))

def prepare_sqlite3():
    global include_dirs
    global sources

    sqlite_dir = os.path.join(root_dir, "sqlite-amalgamation-3470200")
    sqlite_file = os.path.join(root_dir, "sqlite-amalgamation-3470200.zip")
    url = "https://www.sqlite.org/2024/sqlite-amalgamation-3470200.zip"

    if not os.path.exists(sqlite_dir):
        if not os.path.isfile(sqlite_file):
            urllib.request.urlretrieve(url, sqlite_file)

        with zipfile.ZipFile(sqlite_file) as _zip:
            _zip.extractall()

    include_dirs.append(sqlite_dir)
    sources.append(os.path.join(sqlite_dir, 'sqlite3.c'))

def prepare_indexed_gzip():
    global include_dirs
    global sources

    igzip_dir = os.path.join(root_dir, "indexed_gzip-1.10.3", "indexed_gzip")
    igzip_file = os.path.join(root_dir, "indexed_gzip-1.10.3.zip")
    url = "https://github.com/pauldmccarthy/indexed_gzip/archive/refs/tags/v1.10.3.zip"

    if not os.path.exists(igzip_dir):
        if not os.path.isfile(igzip_file):
            urllib.request.urlretrieve(url, igzip_file)

        with zipfile.ZipFile(igzip_file) as _zip:
            _zip.extractall()

    include_dirs.append(igzip_dir)
    sources.extend(glob.glob(os.path.join(igzip_dir, 'zran*.c')))


if sys.platform.startswith('win'):
    comp_args.extend([
        '/D_LFS64_LARGEFILE',
        '/D_LARGEFILE64_SOURCE',
        '/D_FILE_OFFSET_BITS=64'
    ])
else:
    comp_args.extend([
        '-Wno-unused-result',
        '-D_FILE_OFFSET_BITS=64'
    ])

    if sys.platform.startswith('darwin'):
        comp_args.append('-DHAVE_UNISTD_H')
    else:
        comp_args.extend([
            '-D_LFS64_LARGEFILE',
            '-D_LARGEFILE64_SOURCE'
        ])

class CustomBuildExt(build_ext.build_ext):
    def build_extensions(self):
        if not sys.platform.startswith(('win', 'darwin')):
            if '--debug' in sys.argv:
                self.compiler.compiler_so = [
                    opt
                    for opt in self.compiler.compiler_so
                    if opt != '-O2'
                ]
                self.compiler.compiler_so.append('-O0')
                self.compiler.compiler_so.append('--coverage')

                self.compiler.linker_so = [
                    opt
                    for opt in self.compiler.linker_so
                    if opt != '-O2'
                ]
                self.compiler.linker_so.append('-O0')
                self.compiler.linker_so.append('--coverage')

            else:
                self.compiler.compiler_so = [
                    opt
                    for opt in self.compiler.compiler_so
                    if opt != '-g'
                ]

                self.compiler.linker_so = [
                    opt
                    for opt in self.compiler.linker_so
                    if opt != '-g'
                ]

        super().build_extensions()

#if not sys.platform.startswith('linux'):
prepare_sqlite3()
prepare_zlib()
prepare_indexed_gzip()

extension = Extension('pyfastx',
    sources = sources,
    include_dirs = include_dirs,
    extra_compile_args = comp_args,
    extra_link_args = link_args
)

description = (
    "pyfastx is a python module for fast random "
    "access to sequences from plain and gzipped "
    "FASTA/Q file"
)

with open(os.path.join(root_dir, 'README.rst')) as fh:
    long_description = fh.read()

with open(os.path.join(root_dir, 'src', 'version.h')) as fh:
    version = fh.read().split()[2].strip('"')

setup(
    name = 'pyfastx',
    version = version,
    ext_modules = [extension],
    description = description,
    long_description = long_description,
    #long_description_content_type = 'text/x-rst',
    author = 'Lianming Du',
    author_email = 'adullb@qq.com',
    url = 'https://github.com/lmdu/pyfastx',
    license = 'MIT',
    keywords = 'fasta fastq sequence bioinformatics',
    classifiers = [
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "Natural Language :: English",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: C",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: 3.13",
            "Programming Language :: Python :: 3.14",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX :: Linux",
            "Operating System :: Unix",
            "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
    entry_points = {
        'console_scripts': ['pyfastx = pyfastxcli:main']
    },
    py_modules = ["pyfastxcli"],
    test_suite = "tests",
    cmdclass = {
        'build_ext': CustomBuildExt
    }
)



import os
import random
import pyfastx
import pyfaidx
import unittest

join = os.path.join
data_dir = join(os.path.dirname(os.path.abspath(__file__)), 'data')

gzip_fasta = join(data_dir, 'test.fa.gz')
flat_fasta = join(data_dir, 'test.fa')

class IdentifierTest(unittest.TestCase):
	def setUp(self):
		self.fastx = pyfastx.Fasta(gzip_fasta)

		self.fasta = pyfastx.Fasta(flat_fasta)

		self.faidx = pyfaidx.Fasta(flat_fasta, sequence_always_upper=True)

		self.count = len(self.fastx)

	def tearDown(self):
		del self.fastx
		del self.fasta
		del self.faidx

		if os.path.exists('{}.fxi'.format(gzip_fasta)):
			os.remove('{}.fxi'.format(gzip_fasta))

		if os.path.exists('{}.fxi'.format(flat_fasta)):
			os.remove('{}.fxi'.format(flat_fasta))

		if os.path.exists('{}.fai'.format(flat_fasta)):
			os.remove('{}.fai'.format(flat_fasta))

	def get_random_index(self):
		return random.randint(0, self.count-1)

	def test_key_repr(self):
		ids = self.fastx.keys()
		self.assertEqual(repr(ids), "<FastaKeys> contains {} keys".format(len(ids)))

	def test_key_identifier(self):
		fikeys = list(self.faidx.keys())
		fxkeys = list(self.fastx.keys())
		keyobj = self.fastx.keys()

		self.assertEqual(sorted(fikeys), sorted(fxkeys))

		#id counts
		self.assertEqual(len(fikeys), len(keyobj))

		idx = self.get_random_index()
		#get id from identifier class
		self.assertEqual(fikeys[idx], keyobj[idx])

		#negative index
		self.assertEqual(fikeys[idx-len(fikeys)], keyobj[idx-len(fikeys)])
		self.assertEqual(fikeys[-1], keyobj[-1])

		#check contains
		self.assertTrue(self.faidx[idx].name in keyobj)

	def test_key_slice(self):
		fikeys = list(self.faidx.keys())
		fxkeys = self.fastx.keys()
		idx = self.get_random_index()

		#test positive
		self.assertEqual(fikeys[idx:idx+10], fxkeys[idx:idx+10])

		#test negative
		idx = self.get_random_index() - self.count
		self.assertEqual(fikeys[idx:idx+10], fxkeys[idx:idx+10])

	def test_keys_sort(self):
		#sort by id
		keys = self.fastx.keys()

		expect = [seq.name for seq in self.faidx]
		expect.reverse()
		result = [name for name in keys.sort('id', reverse=True)]
		self.assertEqual(expect, result)

		#sort by name
		expect = sorted([seq.name for seq in self.faidx])
		result = [name for name in keys.sort('name')]
		self.assertEqual(expect, result)

		#sort by length
		lens = [(seq.name, len(seq)) for seq in self.faidx]
		expect = [it[0] for it in sorted(lens, key=lambda x: x[1])]
		result = [name for name in keys.sort('length')]
		self.assertEqual(expect, result)

		keys.reset()

	def test_keys_filter(self):
		ids = self.fastx.keys()

		#test greater than
		expect = list(ids.filter(ids>700))
		result = [seq.name for seq in self.faidx if len(seq) > 700]
		self.assertEqual(expect, result)

		#test two compare
		expect = list(ids.filter(600<=ids<=700))
		result = [seq.name for seq in self.faidx if len(seq) >= 600 and len(seq) <= 700]
		self.assertEqual(expect, result)

		#test like compare
		expect = list(ids.filter(ids % 'JZ8226'))
		result = [seq.name for seq in self.faidx if seq.name.startswith('JZ8226')]
		self.assertEqual(expect, result)

		#test all compare
		expect = list(ids.filter(ids % 'JZ8226', ids>=300).sort('name', reverse=True))
		result = [seq.name for seq in self.faidx if seq.name.startswith('JZ8226') and len(seq) >= 300]
		result = sorted(result, reverse=True)
		self.assertEqual(result, expect)

		#get filtered item
		self.assertEqual(result[0], ids[0])
		self.assertEqual(result[-1], ids[-1])

		ids.reset()
		ids.filter(ids % 'JZ8226', ids >= 200)

	def test_id_exception(self):
		keyobj = self.fastx.keys()
		with self.assertRaises(IndexError):
			_ = keyobj[len(keyobj)]

		with self.assertRaises(ValueError):
			_ = keyobj % list

		with self.assertRaises(ValueError):
			_ = keyobj.filter()

		with self.assertRaises(ValueError):
			_ = keyobj.sort('sort')

		with self.assertRaises(ValueError):
			_ = keyobj > list

		with self.assertRaises(TypeError):
			_ = keyobj[list]

if __name__ == '__main__':
	unittest.main()
